/* idhp.c                                                                     */
/* Copyright (c)2009 by  Jie Lian, All Rights Reserved                        */
/* This program simulates the Integrated Dual-chamber Heart & Pacer model.    */
/*  Reference:                                                                */
/*     J Lian, D Muessig, "Heart rhythm and cardiac pacing: an integrated     */
/*     dual-chamber heart and pacer model." Ann Biomed Eng 37(1):64-81, 2009. */
/* Contact J. Lian  jie.lian AT biotronik DOT com)                            */
/*                                                                            */      
/*   This program is free software; you can redistribute it and/or modify     */ 
/*   it under the terms of the GNU General Public License as published by     */ 
/*   the Free Software Foundation; either version 2 of the License, or        */ 
/*   (at your option) any later version.                                      */ 
/*                                                                            */ 
/*   This program is distributed in the hope that it will be useful,          */ 
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of           */ 
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */ 
/*   GNU General Public License for more details.                             */
/*                                                                            */      
/*   You should have received a copy of the GNU General Public License        */
/*   along with this program; if not, write to the Free Software Foundation   */
/*   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA            */ 
/*                                                                            */ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <time.h>    // For random()

#include "idhp.h"    // header file for the IDHP model

#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX_AA 100000 // max number of AA intervals read from external files

// definition of global model variables
EnvStruct  env;     // simulation environment
ArgStruct  arg;     // atrial rhythm generator model
VrgStruct  vrg;     // ventricle rhythm generator model
AtrStruct  atr;     // atrium model
AvjStruct  avj;     // AVJ model
VtrStruct  vtr;     // ventricle model
VeleStruct vele;    // ventricular electrode model
AeleStruct aele;    // atrial electrode model
PacerStruct pacer;  // pacer model

unsigned int rseed; // seed for random number generator, can be passed as an argument
char *MODE[] = {"AOO", "VOO", "AAI", "VVI", "DVI", "DDI", "DDD"};
                    
double aaintv[MAX_AA]; // imported AA intervals (read from external file)

int main(int argc, char *argv[])
{
  time_t utc;

  if (argc<2 || argc>3) {
    fprintf(stdout,"Usage: %s <config_file> [rand_seed]\n",argv[0]);
    fprintf(stdout," rseed = seed for random number generator - specify a long int for repeatability \n\n");
    exit(1);
  }

  ReadConfig(argv[1]);        // read model configuration file
  if (argc>2)
    rseed = atol(argv[2]);    // random seed
  else {                      // use the system clock to give it a random number 
    time(&utc);
    rseed = (long int) utc;
  }

  ModelInit();                // initialize model parameters

  // Main loop, keep running until MAX_BEAT (RR) or MAX_TIME timeout
  while ((env.beatV<env.MAX_BEAT) && (env.t<env.MAX_TIME)) {

    UpdateAtTs();             // model update at sampling time

    if (atr.nA>0 && atr.nR>0 && atr.tA[0]/atr.AntDly+atr.tR[0]/atr.RetDly<=1.0)
      AtrFusion();            // detection of atrial fusion
    if (vtr.nA>0 && vtr.nR>0 && vtr.tA[0]/vtr.AntDly+vtr.tR[0]/vtr.RetDly<=1.0)
      VtrFusion();            // detection of ventricular fusion
    if (pacer.tAP==0 && AtrPace()==CAP)  // timeout for a scheduled Ap and ...
      ArgReset();             // if AP capture, then reset ARG
    else {                    // either no AP, or AP non-capture
      if (unirand()<arg.NZ_Prob)// incidence of atrial noise
        AtrSense(NOISE);      // atrial noise sense
      if (arg.t==0) 
        AtrSense(ArgOutput());// ARG output an atrial impulse, causing atrial sense
      if (atr.nR>0 && atr.tR[0]==0) {
        ArgReset();           // ARG is reset by retrograde P wave
        AtrSense(CONDUCTED);  // retrograde conducted atrial sense
      }
    } // in rare case of simultaneous AtrPace()==LOC & AtrSense(), aele.event = E_ASBLK
    if (pacer.tVP==0 && VtrPace()==CAP)  // timeout for a scheduled Vp and ...
        VrgReset();           // if VP capture, then reset VRG
    else {                    // either no VP, or VP non-capture
      if (unirand()<vrg.NZ_Prob)// incidence of ventricle noise
        VtrSense(NOISE);      // ventricle noise sense
      if (vrg.t==0) 
        VtrSense(VrgOutput());// VRG output an escape impulse, causing ventricle sense
      if (vtr.nA>0 && vtr.tA[0]==0) {
        VrgReset();           // VRG is reset by conducted VS
        VtrSense(CONDUCTED);  // antegrade conducted ventricle sense
      }
    } // in rare case of simultaneous VtrPace()==LOC & VtrSense(), vele.event = E_VSBLK
    if (atr.nA>0 && atr.tA[0]==0)
      AnteHitAvj();           // antegrade atrial impluse hits AVJ
    if (vtr.nR>0 && vtr.tR[0]==0)
      RetrHitAvj();           // VP-induced retrograde wave hits AVJ
    if (avj.nA>0 && avj.tA[0]==0)
      AnteEscAvj();           // ante AV delay timeout and invade vtr
    if (avj.nR>0 && avj.tR[0]==0)
      RetrEscAvj();           // retr AV delay timeout and invade atr

    if (avj.phase==PHASE4) {  // check if AVJ depol & starts ref
      if (avj.Vm>=avj.Vt) {
        ActivateAvj();
        StartAvjRef();
      }
    }
    else {                    // check if AVJ ref ends & return to phase 4
      if (avj.t>=avj.ref)
        StartAvjPh4();
    }
  }

  ModelExit();
} // end main() 

/*********************************************************************
ErrMsg: display error message and abort simulation
Input: msg - message to be displayed
       errcode - exit error code
*********************************************************************/
void ErrMsg(char *msg, int errcode)
{
  fprintf(stderr,"%s\n",msg);
  exit(errcode);
}

/*********************************************************************
RecordMarker: register pacer markers, timers, pace/sense status
Input: fp - marker file pointer
       event - pacer detected event type (incl. blanked sense)
       type  - true sense type or pace CAP/LOC status
*********************************************************************/
void RecordMarker(FILE *fp, int event, int type)
{
  fprintf(fp,"%8.3lf%3d", env.t, event);               // timestamp & event type
  fprintf(fp,"%7.3lf%7.3lf", pacer.tAP, pacer.tVP);    // tAP & tVP
  fprintf(fp,"%7.3lf%7.3lf", pacer.tAB, pacer.tARP);   // tAB & tARP
  fprintf(fp,"%7.3lf%7.3lf", pacer.tFFP, pacer.tPVARP);// tFFP & tPVARP
  fprintf(fp,"%7.3lf", pacer.tSW);                     // tSW
  fprintf(fp,"%7.3lf%7.3lf", pacer.tVB, pacer.tVRP);   // tVB & tVRP
  fprintf(fp,"%7.3lf%2d", pacer.tUTI, pacer.WKBseq);   // tUTI & WKBseq
  fprintf(fp,"%7.3lf", pacer.AVD);                     // AVD
  fprintf(fp,"%7.3lf%7.3lf", pacer.LastAA, pacer.LastAV); // AA & AV intervals
  fprintf(fp,"%7.3lf%7.3lf", pacer.LastVV, pacer.LastVA); // VV & VA intervals
  if (event == E_AP+NUM_VTYPE) {                       // for AP event
    if (type == CAP) fprintf(fp,"%5.1lf\n", pacer.AP_AMP); // CAP: true ApAmp
    else fprintf(fp,"%5.1lf\n",-pacer.AP_AMP);         // LOC: mark (-)ApAmp
  }
  else if (event == E_VP) {                            // for VP event
    if (type == CAP) fprintf(fp,"%5.1lf\n", pacer.VP_AMP); // CAP: true VpAmp
    else fprintf(fp,"%5.1lf\n",-pacer.VP_AMP);         // LOC: mark (-)ApAmp
  }
  else fprintf(fp,"%3d\n",type);                       // mark sense type
}

/*********************************************************************
LogEvent: log event time and type
Input: fp - log file pointer
       msg - message to be logged      
*********************************************************************/
void LogEvent(FILE *fp, char *msg)
{
  fprintf(fp,"T=%.3lf '%s ",env.t,msg);
  if (avj.phase==PHASE4) fprintf(fp,"@PH4' ");
  else fprintf(fp,"@REF' ");
}

/*********************************************************************
LogStat: log statistics after simulation
Input: fp - log file pointer
*********************************************************************/
void LogStat(FILE *fp)
{
	long sum;

  fprintf(fp,"\n----------------------------------------------------\n");
       fprintf(fp,"Random seed = %d\n",rseed);
	fprintf(fp,"Total time (s): %.3lf\n", env.t);
	fprintf(fp,"Total beat: %ld\n", env.beatV);
	sum = vele.cnt[E_VS]+vele.cnt[E_VES]+vele.cnt[E_VRS]+vele.cnt[E_VSBLK];
	fprintf(fp,"Vs Count = %ld [%.2lf%%]\n", 
	       sum, 100.0*sum/env.beatV);
	fprintf(fp,"Vp Count = %ld [%.2lf%%]\n",
	       vele.cnt[E_VP], 100.0*vele.cnt[E_VP]/env.beatV);
	sum = arg.nAA[REFRACTORY]+arg.nAA[PHASE4];
	fprintf(fp,"Af Count = %ld(ref)+%ld(ph4) = %ld\n",
	       arg.nAA[REFRACTORY], arg.nAA[PHASE4], sum);
	fprintf(fp,"PAC=%ld, PVC=%ld, AtrNZ=%ld, VtrNZ=%ld\n",
	       arg.nEB, vrg.nEB, arg.nNZ, vrg.nNZ);
	fprintf(fp,"Mean RR interval = %.3lf <%dppm>\n",
	       env.t/env.beatV, (int)(env.beatV*60/env.t));
	fprintf(fp,"Mean PP interval = %.3lf <%dppm>\n",
	       env.t/env.beatA, (int)(env.beatA*60/env.t));
	sum = vtr.nF[REFRACTORY]+vtr.nF[PHASE4];
	fprintf(fp,"Vtr fusions = %ld(ref)+%ld(ph4) = %ld [%.2lf%%]\n",
	       vtr.nF[REFRACTORY], vtr.nF[PHASE4], sum, 100.0*sum/env.beatV);
	sum = avj.nF[REFRACTORY]+avj.nF[PHASE4];
	fprintf(fp,"AVJ fusions = %ld(ref)+%ld(ph4) = %ld [%.2lf%%]\n",
	       avj.nF[REFRACTORY], avj.nF[PHASE4], sum, 100.0*sum/env.beatV);
	sum = atr.nF[REFRACTORY]+atr.nF[PHASE4];
	fprintf(fp,"Atr fusions = %ld(ref)+%ld(ph4) = %ld [%.2lf%%]\n",
	       atr.nF[REFRACTORY], atr.nF[PHASE4], sum, 100.0*sum/env.beatV);
	fprintf(fp,"Retrgrade Block (w/o fusion) = %ld(avj)+%ld(vtr)\n",
	       avj.nB[RETROGRADE], vtr.nB[RETROGRADE]);
	fprintf(fp,"Antegrade Block (w/o fusion) = %ld(avj)+%ld(vtr)\n",
	       avj.nB[ANTEGRADE],vtr.nB[ANTEGRADE]);
	sum = arg.nRS[REFRACTORY]+arg.nRS[PHASE4];
	fprintf(fp,"ARG Reset Count = %ld(ref)+%ld(ph4) = %ld\n",
	       arg.nRS[REFRACTORY], arg.nRS[PHASE4], sum);
	fprintf(fp,"Active waves = %ld (ante) + %ld (retr)\n",
	      vtr.nA+avj.nA+atr.nA, vtr.nR+avj.nR+atr.nR);
	fprintf(fp,"Ap=%ld, As=%ld, Ars=%ld, AsFfp=%ld, AsPmt=%ld, AsBlk=%ld\n",
	      aele.cnt[E_AP],aele.cnt[E_AS],aele.cnt[E_ARS],
	      aele.cnt[E_ASFFP],aele.cnt[E_ASPMT],aele.cnt[E_ASBLK]);
	fprintf(fp,"Vp=%ld, Vs=%ld, VES=%ld, Vrs=%ld, VsBlk=%ld\n",
	      vele.cnt[E_VP],vele.cnt[E_VS],vele.cnt[E_VES],vele.cnt[E_VRS],vele.cnt[E_VSBLK]);
}

/*********************************************************************
ReadConfig: read configuration file
Input: fn - config filename
*********************************************************************/
void ReadConfig(char *fn)
{
  int i;
  FILE *fp;
  char str[BUFLIM];

  if ((fp=fopen(fn,"rt"))==NULL) ErrMsg("Cannot open config file",1);

  // read parameters of the simulation environment
  ReadLine(fp,-1,env.fnRR);
  ReadLine(fp,-1,env.fnAA);
  ReadLine(fp,-1,env.fnAV);
  ReadLine(fp,-1,env.fnMK);
  ReadLine(fp,-1,env.fnLOG);
  env.MAX_BEAT = (long)ReadLine(fp,500,str);
  env.MAX_TIME = ReadLine(fp,1000,str);
  env.Ts = ReadLine(fp,0.001,str);
  env.RR0 = ReadLine(fp,1.0,str);

  // read parameters of the atrial rhythm generator model
  arg.MODEL = (int)ReadLine(fp,0,str);
  arg.lambda = ReadLine(fp,5,str);
  arg.AAstd = ReadLine(fp,0,str);
  arg.dVmean = ReadLine(fp,33,str);
  arg.dVstd = ReadLine(fp,0,str);
  arg.S1S2 = ReadLine(fp,0.4,str);
  arg.S2S3 = ReadLine(fp,0.4,str);
  arg.EB_Prob = ReadLine(fp,0,str); 
  arg.EB_CI1 = ReadLine(fp,0.2,str);
  arg.EB_CI2 = ReadLine(fp,0.6,str);
  arg.NZ_Prob = ReadLine(fp,0,str); 

  // read parameters of the ventricle rhythm generator model
  vrg.MODEL = (int)ReadLine(fp,2,str);
  vrg.mean = ReadLine(fp,3.0,str);
  vrg.std = ReadLine(fp,0.05,str);
  vrg.EB_Prob = ReadLine(fp,0,str); 
  vrg.EB_CI1 = ReadLine(fp,0.3,str);
  vrg.EB_CI2 = ReadLine(fp,0.8,str);
  vrg.NZ_Prob = ReadLine(fp,0,str); 

  // read parameters of the atrial model
  atr.AntDly = ReadLine(fp,0.03,str);
  atr.RetDly = ReadLine(fp,0.03,str);
  atr.ref = ReadLine(fp,0.05,str);

  // read parameters of the AVJ model
  avj.Vt = ReadLine(fp,-40,str);
  avj.Vr = ReadLine(fp,-90,str);
  avj.dVdt = ReadLine(fp,30,str);
  avj.MinAVDa = ReadLine(fp,0.07,str);
  avj.MinAVDr = ReadLine(fp,0.07,str);
  avj.alpha = ReadLine(fp,0.13,str);
  avj.tau_c = ReadLine(fp,0.1,str);
  avj.MinRef = ReadLine(fp,0.05,str);
  avj.beta = ReadLine(fp,0.25,str);
  avj.tau_r = ReadLine(fp,0.5,str);
  avj.Ref_std = ReadLine(fp,0.0,str);
  avj.delta = ReadLine(fp,10,str);
  avj.theta = ReadLine(fp,10,str);

  // read parameters of the ventricle model
  vtr.AntDly = ReadLine(fp,0.05,str);
  vtr.RetDly = ReadLine(fp,0.15,str);
  vtr.ref = ReadLine(fp,0.1,str);

  // read parameters of the RA electrode model
  aele.ApThresh = ReadLine(fp,1.0,str);

  // read parameters of the RV electrode model
  vele.VpThresh = ReadLine(fp,1.0,str);
  
  // read parameters of the pacer model
  ReadLine(fp,-1,str);
  pacer.MODE = DDD;           // default mode: DDD
  for (i=0;i<strlen(str);i++) str[i] = toupper(str[i]);
  for (i=0;i<NUM_MODE;i++)
    if (strcmp(MODE[i],str)==0) {
      pacer.MODE = i; 
      break; 
    }
  pacer.BI = ReadLine(fp,1.0,str);
  pacer.AVDp = ReadLine(fp,0.15,str);
  pacer.AVDs = ReadLine(fp,0.15,str);
  pacer.PVARP = ReadLine(fp,0.250,str);
  pacer.PVARPext = ReadLine(fp,0.100,str);
  pacer.FFPp = ReadLine(fp,0.15,str);
  pacer.FFPs = ReadLine(fp,0.10,str);
  pacer.ARP = ReadLine(fp,0.300,str);
  pacer.VRP = ReadLine(fp,0.250,str);
  pacer.PVAB = ReadLine(fp,0.03,str);
  pacer.PAVB = ReadLine(fp,0.03,str); 
  pacer.IPAB = ReadLine(fp,0.12,str); 
  pacer.IPVB = ReadLine(fp,0.25,str); 
  pacer.ISAB = ReadLine(fp,0.101,str);
  pacer.ISVB = ReadLine(fp,0.101,str);
  pacer.UTI = ReadLine(fp,0.400,str);
  pacer.SW = ReadLine(fp,0.100,str);
  
  pacer.AP_AMP = ReadLine(fp,3.6,str);
  pacer.VP_AMP = ReadLine(fp,3.6,str);
  pacer.VRS_METHOD = (int)ReadLine(fp,0,str);

  fclose(fp);

//  LogConfig(stdout);    // display model parameters on terminal
}

/*********************************************************************
ReadLine: read a line from the config file
Input:  fp - pointer to the config file
        v  - default value
        str - char pointer (string read after '=' sign)
Output: 'val' read from the line if found, otherwise return 'v'
Note: Comment line starts with '%' sign
      Value is separated from text by '=' 
      Allow only one '=' per line, line width limited by BUFLIM
*********************************************************************/
double ReadLine(FILE *fp, double v, char *str)
{
  char buf[BUFLIM], *p;
  char tok[] = "=";                        // token followed by 'val'
  double val;                              // value read from the line

  *str = '\0';                             // default: null string
  do {
    p = fgets(buf,sizeof(buf),fp);         // read a line
    if (p==NULL) return v;                 // force return on EOF
  } while (strlen(buf)<=1 || buf[0]=='%'); // skip comment or blank lines
  p = strstr(buf,tok);                     // search for "=" sign
  if (p!=NULL) {                           // if found "="
    p += strlen(tok);                      // skip "=" sign
    sscanf(p,"%s",str);                    // read string after "=" sign
    if (sscanf(p,"%lf",&val)>0) return val;// read value after "=" sign
  }                                        // otherwise
  return v;                                // return default value
}

/*********************************************************************
LogConfig: log model configuration
Input: fp - output file pointer
*********************************************************************/
void LogConfig(FILE *fp)
{
  fprintf(fp,"----------------------------------------------------\n");
  fprintf(fp,"*** Simulation Environment ***\n");
  fprintf(fp,"fnRR=%s\nfnAA=%s\nfnAV=%s\nfnMK=%s\nfnLOG=%s\n",
         env.fnRR,env.fnAA,env.fnAV,env.fnMK,env.fnLOG);
  fprintf(fp,"MAX_BEAT=%ld\nMAX_TIME=%.3lf\nTs=%.3lf\nRR0=%.3lf\n",
         env.MAX_BEAT,env.MAX_TIME,env.Ts,env.RR0);
  fprintf(fp,"\n*** Atrial Rhythm Generator Model ***\n");
  fprintf(fp,"MODEL=%d\nlambda=%.3lf\nAAstd=%.3lf\n",
         arg.MODEL,arg.lambda,arg.AAstd);
  fprintf(fp,"dVmean=%.3lf\ndVstd=%.3lf\nS1S2=%.3lf\nS2S3=%.3lf\n",
         arg.dVmean,arg.dVstd,arg.S1S2,arg.S2S3);
  fprintf(fp,"EB_Prob=%.6lf\nEB_CI1=%.3lf\nEB_CI2=%.3lf\nNZ_Prob=%.3e\n",
         arg.EB_Prob,arg.EB_CI1,arg.EB_CI2,arg.NZ_Prob);
  fprintf(fp,"\n*** Ventricle Rhythm Generator Model ***\n");
  fprintf(fp,"MODEL=%d\nmean=%.3lf\nstd=%.3lf\n",vrg.MODEL,vrg.mean,vrg.std);
  fprintf(fp,"EB_Prob=%.6lf\nEB_CI1=%.3lf\nEB_CI2=%.3lf\nNZ_Prob=%.3e\n",
         vrg.EB_Prob,vrg.EB_CI1,vrg.EB_CI2,vrg.NZ_Prob);
  fprintf(fp,"\n*** Atrial Model ***\n");
  fprintf(fp,"AntDly=%.3lf\nRetDly=%.3lf\nref=%.3lf\n",
         atr.AntDly,atr.RetDly,atr.ref); 
  fprintf(fp,"\n*** AVJ Model ***\n");
  fprintf(fp,"Vt=%.3lf\nVr=%.3lf\ndVdt=%.3lf\nMinAVDa=%.3lf\nMinAVDr=%.3lf\n",
         avj.Vt,avj.Vr,avj.dVdt,avj.MinAVDa,avj.MinAVDr);
  fprintf(fp,"alpha=%.3lf\ntau_c=%.3lf\nMinRef=%.3lf\nbeta=%.3lf\n",
         avj.alpha,avj.tau_c,avj.MinRef,avj.beta);
  fprintf(fp,"tau_r=%.3lf\nRef_std=%.3lf\ndelta=%.3lf\ntheta=%.3lf\n",
         avj.tau_r,avj.Ref_std,avj.delta,avj.theta);
  fprintf(fp,"\n*** Ventricle Model ***\n");
  fprintf(fp,"AntDly=%.3lf\nRetDly=%.3lf\nref=%.3lf\n",
         vtr.AntDly,vtr.RetDly,vtr.ref);
  fprintf(fp,"\n*** Atrial Electrode Model ***\n");
  fprintf(fp,"ApThresh=%.3lf\n",aele.ApThresh);
  fprintf(fp,"\n*** Ventricular Electrode Model ***\n");
  fprintf(fp,"VpThresh=%.3lf\n",vele.VpThresh);
  fprintf(fp,"\n*** Pacer Model ***\n");
  fprintf(fp,"MODE=%s\nBI=%.3lf\n",MODE[pacer.MODE],pacer.BI);
  fprintf(fp,"AVDp=%.3lf\nAVDs=%.3lf\n",pacer.AVDp,pacer.AVDs);
  fprintf(fp,"PVARP=%.3lf\nPVARPext=%.3lf\n",pacer.PVARP,pacer.PVARPext);
  fprintf(fp,"FFPp=%.3lf\nFFPs=%.3lf\n",pacer.FFPp,pacer.FFPs);
  fprintf(fp,"ARP=%.3lf\nVRP=%.3lf\n",pacer.ARP,pacer.VRP);
  fprintf(fp,"PVAB=%.3lf\nPAVB=%.3lf\n",pacer.PVAB,pacer.PAVB);
  fprintf(fp,"IPAB=%.3lf\nIPVB=%.3lf\n",pacer.IPAB,pacer.IPVB);
  fprintf(fp,"ISAB=%.3lf\nISVB=%.3lf\n",pacer.ISAB,pacer.ISVB);
  fprintf(fp,"UTI=%.3lf\nSW=%.3lf\n",pacer.UTI,pacer.SW);
  fprintf(fp,"AP_AMP=%.1lf\nVP_AMP=%.1lf\n",pacer.AP_AMP,pacer.VP_AMP);
  fprintf(fp,"VRS_METHOD=%d\n",pacer.VRS_METHOD);
  fprintf(fp,"----------------------------------------------------\n\n");
}

/*********************************************************************
ModelInit: initialization of model parameters
*********************************************************************/
void ModelInit()
{
  int i;
  long aacnt;           // counter of aaintv[]

  if ((env.fprr=fopen(env.fnRR,"wt"))==NULL)  // output file of RR intervals
    ErrMsg("Cannot open output file of RR intervals",1);
  if ((env.fpaa=fopen(env.fnAA,"wt"))==NULL)  // output file of AA intervals
    ErrMsg("Cannot open output file of AA intervals",1);
  if ((env.fpav=fopen(env.fnAV,"wt"))==NULL)  // output file of AVJ status
    ErrMsg("Cannot open output file of AVJ status",1);
  if ((env.fpmk=fopen(env.fnMK,"wt"))==NULL)  // output file of event markers
    ErrMsg("Cannot open output file of event markers",1);
  if ((env.fplog=fopen(env.fnLOG,"wt"))==NULL)// output file of event log
    ErrMsg("Cannot open output file of event log",1);

  env.beatA = env.beatV = 0;               // clear atrial/ventricle beat counters
  avj.phase = PHASE4;                      // set AVJ phase = PHASE4
  avj.Vm = avj.Vr;                         // set Vm to Vr
  avj.V4 = avj.dVdt * env.Ts;              // initialize V4 (mV/ms)
  avj.RT0 = 0;                             // AVJ start recovery\
  UpdateTimer(&env.t, OVW, 0);             // reset simulation time clock
  UpdateTimer(&avj.t, OVW, 0);             // clear AVJ refractory period

// assume input lambda is the AA interval (sec), NOT the rate (pulses per second)
//  if (arg.lambda < 1) arg.lambda = 1.0/arg.lambda;

  avj.nA = avj.nR = 0;                     // no activation wave in AVJ
  vtr.nA = vtr.nR = 0;                     // no activation wave in vtr
  atr.nA = atr.nR = 0;                     // no activation wave in atr
  for (i=0;i<MAX_TM;i++) {                 
    UpdateTimer(&avj.tA[i], OVW, -1);      // disable AVJ ante activation timers
    UpdateTimer(&avj.tR[i], OVW, -1);      // disable AVJ retr activation timers
    UpdateTimer(&vtr.tA[i], OVW, -1);      // disable vtr ante activation timers
    UpdateTimer(&vtr.tR[i], OVW, -1);      // disable vtr retr activation timers
    UpdateTimer(&atr.tA[i], OVW, -1);      // disable atr ante activation timers
    UpdateTimer(&atr.tR[i], OVW, -1);      // disable atr retr activation timers
  }  

  arg.nAA[REFRACTORY] = arg.nAA[PHASE4] = 0;  // clear ARG output counters
  arg.nRS[REFRACTORY] = arg.nRS[PHASE4] = 0;  // clear ARG reset counters
  arg.nEB    = vrg.nEB    = 0;             // clear ARG & VRG ectopic beat counters
  arg.nNZ    = vrg.nNZ    = 0;             // clear ARG & VRG noise counters
  arg.type   = vrg.type   = INTRINSIC;     // init ARG and VRG output types
   
  atr.nF[REFRACTORY] = atr.nF[PHASE4] = 0; // clear atrial fusion counters
  avj.nF[REFRACTORY] = avj.nF[PHASE4] = 0; // clear AVJ fusion counters
  vtr.nF[REFRACTORY] = vtr.nF[PHASE4] = 0; // clear ventricle fusion counters
  avj.nB[ANTEGRADE] = avj.nB[RETROGRADE] = 0; // clear AVJ block counters
  vtr.nB[ANTEGRADE] = vtr.nB[RETROGRADE] = 0; // clear ventricle block counters
  atr.nB[ANTEGRADE] = atr.nB[RETROGRADE] = 0; // clear atrial block counters
  for (i=0;i<NUM_VTYPE;i++) vele.cnt[i] = 0; // clear vele event counters
  for (i=0;i<NUM_ATYPE;i++) aele.cnt[i] = 0; // clear aele event counters

  avj.tA0 = avj.tR0 = 0;                   // clear ante/retr AVD timeout time
  aele.tA0 = vele.tR0 = 0;                 // clear last A/V activation time
  
  PacerInit();                             // initialize pacer parameters
  
  srandom(rseed);  // Initialize randomization with seeded random number

  if (arg.MODEL == 6) aacnt = ReadAAintv(aaintv);  // read aaintv[] from external file
    
//  UpdateTimer(&arg.t, SET, GetAtrIntv());  // Get 1st interval for next ARG output
//  UpdateTimer(&vrg.t, SET, GetVtrIntv());  // Get 1st interval for next VRG output
  UpdateTimer(&arg.t, SET, GetNextIntv(ATRIUM, &arg.type));   // Get 1st ARG output interval
  UpdateTimer(&vrg.t, SET, GetNextIntv(VENTRICLE, &vrg.type));// Get 1st VRG output interval
  
  LogConfig(env.fplog);                    // log config parameters
}

/*********************************************************************
PacerInit: initialization of pacer parameters
*********************************************************************/
void PacerInit()
{
  if (pacer.BI<pacer.UTI) pacer.BI = pacer.UTI; // ensure BI >= UTI
  if (pacer.AVDp<pacer.AVDs) pacer.AVDs = pacer.AVDp; // ensure AVDp >= AVDs
  if (pacer.MODE==DDD || pacer.MODE==DDI || pacer.MODE==DVI)
    pacer.AVD = pacer.AVDp;                 // init AVD to AVDp
  if ((pacer.MODE==DDD || pacer.MODE==DDI) && pacer.ARP<pacer.AVD)
    pacer.ARP = pacer.AVD;                  //  Auto ARP (DDD & DDI)
  pacer.WKBseq = 0;                         // init null WKB sequence

  // by default clear all pacer timers
  UpdateTimer(&pacer.tAB, OVW, -1);         // default: no AB window
  UpdateTimer(&pacer.tARP, OVW, -1);        // default: no ARP window
  UpdateTimer(&pacer.tVB, OVW, -1);         // default: no VB window
  UpdateTimer(&pacer.tVRP, OVW, -1);        // default: no VRP window
  UpdateTimer(&pacer.tFFP, OVW, -1);        // default: no FFP window
  UpdateTimer(&pacer.tPVARP, OVW, -1);      // default: no PVARP window
  UpdateTimer(&pacer.tUTI, OVW, -1);        // default: no UTI window
  UpdateTimer(&pacer.tAVD, OVW, -1);        // default: no AVD window
  UpdateTimer(&pacer.tSW, OVW, -1);         // default: no SW window
  UpdateTimer(&pacer.tAP, OVW, -1);         // default: disable AP
  UpdateTimer(&pacer.tVP, OVW, -1);         // default: disable VP
  UpdateTimer(&pacer.tUsedA, OVW, -1);      // default: disable AV & AA measurement
  UpdateTimer(&pacer.tUsedV, OVW, -1);      // default: disable VA & VV measurement

  // initialize pacer timers: for AAI & AOO modes 
  if (pacer.MODE==AAI || pacer.MODE==AOO) { // modes AAI & AOO
    pacer.LastUsedA = aele.event = E_AP;    // assuming starts with an AP event
    UpdateTimer(&pacer.tAP, SET, pacer.BI); // start BI for AP
    UpdateTimer(&pacer.tUsedA, OVW, 0);     // start AA interval measurement
    if (pacer.MODE==AAI) {                  // mode AAI (w/ AS)
      UpdateTimer(&pacer.tAB, SET, pacer.IPAB); // start IPAB
      UpdateTimer(&pacer.tARP, SET, pacer.ARP); // start ARP
    }                                       // otherwise AOO (no AS)
  }
  else {      // for modes DDD, DDI, VVI, VOO, DVI
    pacer.LastUsedV = vele.event = E_VP;    // assuming starts with a VP event
    if (pacer.MODE==DDD || pacer.MODE==DDI || pacer.MODE==DVI)
      UpdateTimer(&pacer.tUTI, SET, pacer.UTI);         // start UTI
    UpdateTimer(&pacer.tUsedV, OVW, 0);     // start VV and/or VA interval measurement
    if (pacer.MODE==DDD)                                // A-based timing: DDD
      UpdateTimer(&pacer.tAP, SET, pacer.BI-pacer.AVD); // continue BI (post AVDp) for AP
    else if (pacer.MODE==VOO)                           // V-based timing: VOO
      UpdateTimer(&pacer.tVP, SET, pacer.BI);           // start BI for VP
    else {                                              // V-based timing: DDI, VVI, DVI
      pacer.NextVV = GetVpIntv(env.RR0,pacer.BI,pacer.NextVV,pacer.VRS_METHOD);
      UpdateTimer(&pacer.tVP, SET, pacer.NextVV);       // start VV timer for VP
      if (pacer.MODE==DDI || pacer.MODE==DVI) {
        pacer.NextVA = pacer.NextVV-pacer.AVD;  
        UpdateTimer(&pacer.tAP, SET, pacer.NextVA);     // start VA timer for AP        
      }
    }
    if (pacer.MODE != VOO) {                         // modes w/ VS: DDD, DDI, VVI, DVI 
      UpdateTimer(&pacer.tVB, SET, pacer.IPVB);      // start IPVB
      UpdateTimer(&pacer.tVRP, SET, pacer.VRP);      // start VRP
    }
    if (pacer.MODE==DDD || pacer.MODE==DDI) {        // modes w/ VP & AS
      UpdateTimer(&pacer.tAB, SET, pacer.PVAB);      // start PVAB
      UpdateTimer(&pacer.tFFP, SET, pacer.FFPp);     // start FFPp
      UpdateTimer(&pacer.tPVARP, SET, pacer.PVARP);  // start PVARP
    }
  }
}

/*********************************************************************
ReadLine: read AA intervals from external file,
return vector of AA intervals (aaintv[]) and its count 
*********************************************************************/
long ReadAAintv(double *aaintv)
{
  FILE *fp;
  char buf[BUFLIM], *p;
  double aa;
  long aacnt;           // counter of aaintv[]
    
  aacnt = 0;
  printf("External AA interval file: ");
  scanf("%s",buf);
  if ((fp=fopen(buf,"rt"))==NULL) 
    ErrMsg("Cannot open external AA interval file to read",1);
  for (;;) {
     p = fgets(buf,sizeof(buf),fp);      // read a line
     if (p==NULL) break;                 // force return on EOF
     if (sscanf(p,"%lf",&aa)>0) {
       aaintv[aacnt] = aa;
       aacnt++;
       if (aacnt > MAX_AA) break;
     }     
  }                            
  fclose(fp);
  return (aacnt);
}

/*********************************************************************
ModelExit: exit simulation
*********************************************************************/
void ModelExit()
{
  LogStat(stdout);                   // output statistics to stdout
  LogStat(env.fplog);                // output statistics to log file
  fclose(env.fprr);
  fclose(env.fpaa);
  fclose(env.fpav);
  fclose(env.fpmk);
  fclose(env.fplog);
}

/*********************************************************************
UpdateAtTs: model update at sampling time
*********************************************************************/
void UpdateAtTs()
{
  int i;

  UpdateTimer(&env.t, INC, env.Ts); // update (increment) simulation timer

  UpdateTimer(&arg.t, DEC, env.Ts);    // decrement ARG timer
  UpdateTimer(&vrg.t, DEC, env.Ts);    // decrement VRG timer

  // update AVJ and ventricle activation timers
  for (i=0;i<avj.nA;i++) UpdateTimer(&avj.tA[i], DEC, env.Ts);
  for (i=0;i<avj.nR;i++) UpdateTimer(&avj.tR[i], DEC, env.Ts);
  for (i=0;i<vtr.nA;i++) UpdateTimer(&vtr.tA[i], DEC, env.Ts);
  for (i=0;i<vtr.nR;i++) UpdateTimer(&vtr.tR[i], DEC, env.Ts);
  for (i=0;i<atr.nA;i++) UpdateTimer(&atr.tA[i], DEC, env.Ts);
  for (i=0;i<atr.nR;i++) UpdateTimer(&atr.tR[i], DEC, env.Ts);

  // update pacer timers
  UpdateTimer(&pacer.tAVD, DEC, env.Ts);   
  UpdateTimer(&pacer.tPVARP, DEC, env.Ts);   
  UpdateTimer(&pacer.tFFP, DEC, env.Ts);   
  UpdateTimer(&pacer.tARP, DEC, env.Ts);   
  UpdateTimer(&pacer.tVRP, DEC, env.Ts);   
  UpdateTimer(&pacer.tAB, DEC, env.Ts);   
  UpdateTimer(&pacer.tVB, DEC, env.Ts);   
  UpdateTimer(&pacer.tUTI, DEC, env.Ts);  
  UpdateTimer(&pacer.tSW, DEC, env.Ts);  
  UpdateTimer(&pacer.tVP, DEC, env.Ts);   // update VP timer
  UpdateTimer(&pacer.tAP, DEC, env.Ts);   // update AP timer
  UpdateTimer(&pacer.tUsedA, INC, env.Ts); // update AA/AV measurement
  UpdateTimer(&pacer.tUsedV, INC, env.Ts); // update VA/VV measurement
  
  if (avj.phase==PHASE4) {        // during phase 4
    avj.Vm += avj.V4;             // Vm linear rise from Vr to Vt
    if (avj.Vm>=avj.Vt) avj.dir = ANTEGRADE;
  }
  else                            // during refractory period
    UpdateTimer(&avj.t, INC, env.Ts); // update AVJ refractory timer
}

/*********************************************************************
UpdateTimer: update pacer timer
args:  tm - pointer of the timer
       action - OVW, SET, INC, or DEC
       val - overwrite/set value (OVW, SET) or inc/dec amount (INC, DEC)
*********************************************************************/
void UpdateTimer(double *tm, int action, double val)
{
  if (action == OVW) *tm = val;     // override timer w/o condition
  else if (action == SET) {         // set (start) timer w/ condition that 
    if (*tm < val) *tm = val;       // not override longer running timer
  }
  else if (*tm >= 0) {              // INC or DEC for active timer only
    if (action == INC) *tm += val;  // INC timer
    else {                          // DEC timer
      *tm -= val;
      if (*tm < 0) *tm = 0;         // time out
    }
  }                                 // no action for inactive timer
}

/*********************************************************************
AnteHitAvj: event service when antegrade impulse hits AVJ
*********************************************************************/
void AnteHitAvj()
{
  double ext;
  long r;

  LogEvent(env.fplog,"AnteHitAvj"); 
  
  atr.nA = StopWave(atr.tA, atr.nA);  // stop antegrade atrial timer   
  if (avj.phase==PHASE4) {  // AVJ in phase IV
    avj.Vm += arg.dV;       // delta increment of Vm
    if (avj.Vm>=avj.Vt) avj.dir = ANTEGRADE;
    fprintf(env.fplog,      // log after AnteHitAvj() service @PH4
      "Vm=%.2lf+%.2lf=%.2lf, atr.nA=%d\n",
      avj.Vm-arg.dV, arg.dV, avj.Vm, atr.nA);
  }
  else {                    // AVJ in refractory period
    r = arg.dV/(avj.Vt-avj.Vr); // relative strength of the AF impulse
    if (r>1) r = 1.0;       // with upper limit 1.0
    if (avj.nR>0 && r==1.0) { // presence of retr wave in AVJ & strong AF
      avj.nR = StopWave(avj.tR, avj.nR);  // stop the retr AVJ timer
      avj.nF[avj.phase]++;    // incr AVJ fusion counter
    } // TBD: cumulative effect of weak AF bombardment in AVJ refractory period?
    else avj.nB[ANTEGRADE]++; // incr counter of blocked ante waves (w/o AVJ fusion)
    r = pow(avj.t/avj.ref,avj.theta)*pow(r,avj.delta);
    ext = avj.MinRef*r;
    avj.ref += ext;         // ref extension due to concealed AF
    fprintf(env.fplog,      // log after AnteHitAvj() service @REF
      "ref=%.3lf+%.3lf=%.3lf, atr.nA=%d, avj.nR=%d, nB(a)=%d, nF(R)=%d\n",
      avj.ref-ext, ext, avj.ref, atr.nA, avj.nR,avj.nB[ANTEGRADE],avj.nF[REFRACTORY]);
  }
}

/*********************************************************************
RetrHitAvj: event service when retrograde impulse hits AVJ
*********************************************************************/
void RetrHitAvj()
{
  int i;
  double r,ext;

  LogEvent(env.fplog,"RetrHitAvj"); 

  vtr.nR = StopWave(vtr.tR, vtr.nR);   // stop a retr vtr activation timer
  
  if (avj.phase==PHASE4) {             // retr invasion in phase 4
    if (avj.Vm<avj.Vt) {               // retr AVJ depolarization
      avj.Vm = avj.Vt;                 // suprathreshold AVJ immediately
      avj.dir = RETROGRADE;            // retrograde activation
      fprintf(env.fplog,               // log after RetHitAvj() service @PH4
        "vtr.nR=%d, Vm=%.2lf Retr AVJ activation\n", vtr.nR, avj.Vm);
    }
    else {        // simultaneous ante/retr depol (rare), no wave in AVJ
      avj.dir = BIDIRECTION;
      fprintf(env.fplog,               // log after RetHitAvj() service @PH4
        "vtr.nR=%d, Vm=%.2lf Bidir AVJ activation\n", vtr.nR, avj.Vm);
    }
  }
  else {                               // retr invasion in refractory period
    if (avj.nA>0) {                    // presence of ante wave in AVJ
      avj.nA = StopWave(avj.tA, avj.nA);  // stop the ante AVJ timer
      avj.nF[avj.phase]++;             // incr AVJ fusion counter
    }
    else avj.nB[RETROGRADE]++;         // incr counter of blocked retr waves (w/o AVJ fusion)
    r = pow(avj.t/avj.ref,avj.theta);  // strong impulse
    ext = avj.MinRef*r;
    avj.ref += ext;                    // ref extension by concealed impulse
    fprintf(env.fplog,                 // log after RetHitAvj() service @REF
      "ref=%.2lf+%.2lf=%.2lf, vtr.nR=%d, avj.nA=%d, nB(r)=%d, nF(R)=%d\n", 
      avj.ref-ext,ext,avj.ref,vtr.nR,avj.nA,avj.nB[RETROGRADE],avj.nF[REFRACTORY]);
  }
}

/*********************************************************************
ActivateAvj: AVJ depolarization service
*********************************************************************/
void ActivateAvj()
{
  LogEvent(env.fplog,"ActivateAvj"); 

  avj.RT = env.t-avj.RT0;             // calculated AVJ recovery time 
  avj.AntDly = avj.MinAVDa+avj.alpha*exp(-avj.RT/avj.tau_c); // ante AV delay
  avj.RetDly = avj.MinAVDr+avj.alpha*exp(-avj.RT/avj.tau_c); // retr AV delay

  if (avj.dir==ANTEGRADE) {           // antegrade AVJ activation
    if (avj.nR==0) {                  // no existing retr activation in AVJ (expected)
      avj.nA = StartWave(avj.tA, avj.nA, avj.AntDly); // start a new timer for ante AVD
      fprintf(env.fpav,               // record AVJ status
        "%.3lf %d %.3lf %.3lf ", env.t, avj.dir, avj.RT, avj.AntDly);
      fprintf(env.fplog,              // log after ActivateAvj() service
        "start AVDa = %.3lf, avj.nA=%d\n", avj.AntDly, avj.nA);     
    }
    else { // collide w/ existing retr wave (AVJ activation is NOT possible if refractory)
      avj.nR = StopWave(avj.tR, avj.nR); // stop a retr AVJ activation timer
      avj.nF[avj.phase]++;            // incr AVJ collision counter (Phase 4)
      fprintf(env.fpav,               // record AVJ status
        "%.3lf %d %.3lf %d ", env.t, avj.dir, avj.RT, -1);
      fprintf(env.fplog,              // log after ActivateAvj() service
        "collide w/ retr wave, avj.nF(IV)=%ld\n", avj.nF[avj.phase]);
    }
  }
  else if (avj.dir==RETROGRADE) {     // retrograde AVJ activation
    if (avj.nA==0) {                  // no existing ante activation in AVJ (expected)
      avj.nR = StartWave(avj.tR, avj.nR, avj.RetDly);  // start a new timer for retr AVD
      fprintf(env.fpav,               // record AVJ status
        "%.3lf %d %.3lf %.3lf ", env.t, avj.dir, avj.RT, avj.RetDly);
      fprintf(env.fplog,              // log after ActivateAvj() service
        "start AVDr = %.3lf, avj.nR=%d\n", avj.RetDly, avj.nR);     
    }
    else { // collide w/ existing ante wave (AVJ activation is NOT possible if refractory)
      avj.nA = StopWave(avj.tA, avj.nA); // stop an ante AVJ activation timer
      avj.nF[avj.phase]++;            // incr AVJ collision counter (phase 4)
      fprintf(env.fpav,               // record AVJ status
        "%.3lf %d %.3lf %d ", env.t, avj.dir, avj.RT, -1);
      fprintf(env.fplog,              // log after ActivateAvj() service
        "collide w/ ante wave, avj.nF(IV)=%ld\n", avj.nF[avj.phase]);
    }
  }
  else {                              // simultaneous bidirectional depol
    avj.nF[avj.phase]++;              // incr AVJ collision counter (phase 4)
    fprintf(env.fpav,                 // record AVJ status
      "%.3lf %d %.3lf %d ", env.t, avj.dir, avj.RT, -1);
    fprintf(env.fplog,                // log after ActivateAvj() service
      "bidir depolarization, avj.nF(IV)=%ld\n", avj.nF[avj.phase]);
  }  
}

/*********************************************************************
StartAvjRef: AVJ starts refractory period
*********************************************************************/
void StartAvjRef()
{
  // calculate AVJ refractory period
  avj.ref = avj.MinRef+avj.beta*(1-exp(-avj.RT/avj.tau_r));
  // assume AVJ refr period >= AVD 
  // then -- not possible to ActivateAvj before AVD expires
  //         not possible for multi-waves in AVJ (avj.nA<=1, avj.nR<=1)
  if (avj.nA>0 && avj.ref<avj.AntDly) avj.ref = avj.AntDly;  // limited by AVD
  if (avj.nR>0 && avj.ref<avj.RetDly) avj.ref = avj.RetDly;  // limited by AVD
  avj.ref = gaussian(avj.ref,avj.Ref_std,unirand(),unirand());
  avj.phase = REFRACTORY;             // update AVJ phase to REFRACTORY
  UpdateTimer(&avj.t, OVW, 0);        // start AVJ refractory timer
  
  fprintf(env.fpav,"%.3lf\n", avj.ref); // record AVJ status
  fprintf(env.fplog,                  // log after StartAvjRef() service
    "T=%.3lf 'Start AVJ REF' RT=%.3lf, ref=%.3lf\n", 
    env.t, avj.RT, avj.ref);  
}

/*********************************************************************
StartAvjPh4: AVJ starts phase 4
*********************************************************************/
void StartAvjPh4()
{
  avj.phase = PHASE4;                 // update AVJ phase to PHASE4
  avj.Vm = avj.Vr;                    // set Vm to Vr
  avj.RT0 = env.t;                    // AVJ starts recovery

  fprintf(env.fplog,                  // log after StartAvjPh4() service
    "T=%.3lf 'Start AVJ PH4' Vm=%.2lf, avj.nA=%d, avj.nR=%d\n",
    env.t, avj.Vm, avj.nA, avj.nR);  
}

/*********************************************************************
StartWave: start a new activation timer in atr, vtr, or avj
Input: tm - pointer of the timers (return updated timers)
       n  - number of active timers (return incremented timer count)
       delay - timer delay to be set
*********************************************************************/
int StartWave(double *tm, int n, double delay)
{
  tm[n++] = delay; // start a new delay timer, then incr activation counter
  if (n > MAX_TM) ErrMsg("Too many activation waves!",2);
  return n;
}

/*********************************************************************
StopWave: stop an activation timer in atr, vtr, or avj
Input: tm - pointer of the timers (return updated timers)
       n  - number of active timers (return decremented timer count)
*********************************************************************/
int StopWave(double *tm, int n)
{
  int i;

  for (i=0;i<n-1;i++) tm[i] = tm[i+1]; // update activation queue
  tm[--n] = -1; // first decrement timer counter, then disable a timer
  return n;
}
   
/*********************************************************************
AnteEscAvj: event service when ante wave escapes AVJ and invade vtr
*********************************************************************/
void AnteEscAvj()
{
  LogEvent(env.fplog,"AnteEscAvj"); 
  
  avj.nA = StopWave(avj.tA, avj.nA); // stop an ante AVJ activation timer
  if (avj.nA==0) avj.RT0 = env.t;    // AVJ starts recovery

  if (env.t-avj.tA0<vtr.ref) {       // non-capture because vtr refr
    vtr.nB[ANTEGRADE]++;             // will not generate ante wave in vtr
    fprintf(env.fplog,               // log after AnteEscAvj() service
      "avj.nA=%d, vtr.nB(a)=%d(V-Blk)\n", avj.nA, vtr.nB[ANTEGRADE]);
    return;                          // force return
  }                                  // otherwise, capture vtr and ...
  vtr.nA = StartWave(vtr.tA, vtr.nA, vtr.AntDly); // starts an ante vtr activation
  avj.tA0 = env.t;                   // update last ante AVD expiration time
  
  fprintf(env.fplog,                 // log after AnteEscAvj() service
    "avj.nA=%d, vtr.nA=%d\n", avj.nA, vtr.nA);
}

/*********************************************************************
RetrEscAvj: event service when retr wave escapes AVJ and invade atrium
*********************************************************************/
void RetrEscAvj()
{
  LogEvent(env.fplog,"RetrEscAvj"); 
  
  avj.nR = StopWave(avj.tR, avj.nR); // stop a retr AVJ activation timer
  if (avj.nR==0) avj.RT0 = env.t;    // AVJ starts recovery

  if (env.t-avj.tR0<atr.ref) {       // non-capture because atr refr
    atr.nB[RETROGRADE]++;            // will not generate retr wave in atr
    fprintf(env.fplog,               // log after RetrEscAvj() service
      "avj.nR=%d, atr.nB(r)=%d\n", avj.nR, atr.nB[RETROGRADE]);
    return;                          // force return
  }                                  // otherwise, capture atr and ...
  atr.nR = StartWave(atr.tR, atr.nR, atr.RetDly); // starts a retr atr activation
  avj.tR0 = env.t;                   // update last retr AVD expiration time

  fprintf(env.fplog,                 // log after RetrEscAvj() service
    "avj.nR=%d, atr.nR=%d\n",avj.nR, atr.nR); 
}

/*********************************************************************
ArgOutput: event service when ARG outputs an atrial impuse
*********************************************************************/
int ArgOutput()
{
  int type = arg.type;               // keep a copy of ARG output type
  
  LogEvent(env.fplog,"ArgOutput"); 
  
  if (type==ECTOPIC) {               // AES 
    arg.dV = avj.Vt-avj.Vr;          // assume strong AES impulse
    arg.nEB++;                       // increment AES counter
    fprintf(env.fplog,"AES#%ld ", arg.nEB);
  }
  else {                             // non-AES, normal atril depol.
    arg.dV = gaussian(arg.dVmean,arg.dVstd,unirand(),unirand()); // AF strength
    if (arg.dV<0) arg.dV = 0;        // limit dV >= 0
  }

  // Get the random interval for next atrial depolarization
//  arg.t = GetAtrIntv();              // reset ARG timer
  arg.t = GetNextIntv(ATRIUM, &arg.type);  // reset ARG timer
  if (env.t-aele.tA0 >= atr.ref)     // start ante atr wave if atrial non-refr 
    atr.nA = StartWave(atr.tA, atr.nA, atr.AntDly); //otherwise, no new wave
  arg.nAA[avj.phase]++;              // increment atrial impulse counter
    
  fprintf(env.fplog,                 // log after ArgOutput() service
    "arg.dV=%.3lf, arg.nAA[%d]=%d, atr.nA=%d, NextAA=%.3lf\n", 
    arg.dV, avj.phase, arg.nAA[avj.phase], atr.nA, arg.t);
    
  return type;
}

/*********************************************************************
VrgOutput: event service when VRG outputs an escape impuse
*********************************************************************/
int VrgOutput()
{
  int type = vrg.type;             // keep a copy of VRG output type
  
  LogEvent(env.fplog,"VrgOutput"); 
  
  vrg.nEB++;                       // increment VES counter

//  vrg.t = GetVtrIntv();            // reset VRG timer
  vrg.t = GetNextIntv(VENTRICLE, &vrg.type);  // reset VRG timer
  if (env.t-vele.tR0 >= vtr.ref)   // start retr vtr wave if vtr non-refr 
    vtr.nR = StartWave(vtr.tR, vtr.nR, vtr.RetDly); // otherwise no new wave
    
  fprintf(env.fplog,               // log after VrgOutput() service
    "PVC#%ld, vtr.nR=%d, NextVES=%.3lf\n", vrg.nEB, vtr.nR, vrg.t);
    
  return type;
}

/*********************************************************************
ArgReset: event service to reset ARG 
*********************************************************************/
void ArgReset()
{
  LogEvent(env.fplog,"ArgReset"); 

  arg.nRS[avj.phase]++;              // increment ARG reset counter
  // Get the random interval for next atrial depolarization
//  arg.t = GetAtrIntv();              // reset ARG timer
  arg.t = GetNextIntv(ATRIUM, &arg.type); // reset ARG timer
  
  fprintf(env.fplog,                 // log after ArgReset() service
    "arg.nRS[%d]=%d, atr.nR=%d, NextAA=%.3lf\n", 
    avj.phase, arg.nRS[avj.phase], atr.nR, arg.t);  
}

/*********************************************************************
VrgReset: event service to reset VRG 
*********************************************************************/
void VrgReset()
{
//  vrg.t = GetVtrIntv();              // reset ARG timer
  vrg.t = GetNextIntv(VENTRICLE, &vrg.type); // reset ARG timer
}

/*********************************************************************
AtrFusion: event service when detecting atrial fusion
*********************************************************************/
void AtrFusion()
{
  LogEvent(env.fplog,"AtrFusion"); 

  atr.nF[avj.phase]++;               // incr atrial fusion counter
  atr.nA = StopWave(atr.tA, atr.nA); // stop an ante atr activation timer
  atr.nR = StopWave(atr.tR, atr.nR); // stop a retr atr activation timer
  
  fprintf(env.fplog,                 // log after AtrFusion() service
    "atr.nF(%d)=%ld, atr.nA=%d, atr.nR=%d\n", 
    avj.phase, atr.nF[avj.phase], atr.nA, atr.nR);
}

/*********************************************************************
VtrFusion: event service when detecting ventricular fusion
*********************************************************************/
void VtrFusion()
{
  LogEvent(env.fplog,"VtrFusion"); 

  vtr.nF[avj.phase]++;               // incr ventricular fusion counter
  vtr.nA = StopWave(vtr.tA, vtr.nA); // stop an ante vtr activation timer
  vtr.nR = StopWave(vtr.tR, vtr.nR); // stop a retr vtr activation timer
  
  fprintf(env.fplog,                 // log after VtrFusion() service
    "vtr.nF(%d)=%ld, vtr.nA=%d, vtr.nR=%d\n", 
    avj.phase, vtr.nF[avj.phase], vtr.nA, vtr.nR);
}

/*********************************************************************
RegBeat:  register atrial/ventricle beat and PP/RR interval
args:  beat - point to beat counter (return updated beat counter)
       tm0  - last activation time (return updated activation time)
       event - event type
       fp - point to output file         
*********************************************************************/
double RegBeat(long *beat, double *tm0, int event, FILE *fp)
{
  double intv;
  
  (*beat)++;                 // increment beat counter
  intv = env.t-(*tm0);       // record PP interval or RR interval
  *tm0 = env.t;              // update last capture time (tissue start refr.)
  fprintf(fp, "%.3lf %d\n", intv, event); // record PP/RR interval and event type
  return intv;
}

/*********************************************************************
VtrSense: event service for ventricular sense
args: type - sense type (INSTRINSIC, ECTOPIC, NOISE, CONDUCTED)
*********************************************************************/
void VtrSense(int type)
{
  char *vevt[] = {"VP", "VS", "VES", "VRS", "VSBLK"};

  LogEvent(env.fplog,"VtrSense"); 

  if (type == CONDUCTED)                       // conducted Vs: stop antegrade vtr wave
    vtr.nA = StopWave(vtr.tR, vtr.nA);        
  else if (type == NOISE) {                    // noise Vs: increment noise VS counter
    vrg.nNZ++;
    fprintf(env.fplog, "VtrNoise #%ld ", vrg.nNZ);
  }                                            // otherwise: VRG output, no action
  
  if (pacer.MODE==AAI || pacer.MODE==AOO || pacer.MODE==VOO) 
    vele.event = E_VSBLK;                      // modes w/o VS, treat as VsBlk
  else {                                       // modes DDD, DDI, VVI, DVI
    if (pacer.tVB>0) vele.event = E_VSBLK;     // VS in blanking window
    else {                                     // unblanked vtr sense: VS, VES, or VRS
      if (pacer.tVRP>0) vele.event = E_VRS;    // VS in vtr refractory window
      else {                                   // tracked VS or VES
        if (pacer.MODE==VVI || pacer.MODE==DVI)// modes w/o AS
          vele.event = E_VS;                   // VS (no VES classification)
        else {                                 // modes DDD & DDI
          if (pacer.tAVD>0) {                  // VS in AVD
            vele.event = E_VS;                 // conducted VS
            if (pacer.tSW>0)                   // safety window VS
              fprintf(env.fplog, "*SW_VS* ");
          }
          else                                 // otherwise, VES
            vele.event = E_VES; 
        }
      }
    }
  }
  vele.cnt[vele.event]++;                      // increment respective VS counter
  PacerInterval(vele.event);                   // update pacer measured intervals
  PacerTimerVx(pacer.MODE, vele.event);        // update pacer timers after VS
  
  RecordMarker(env.fpmk, vele.event, type);    // record ventricle event markers
  if (type != NOISE) 
    vele.RR = RegBeat(&env.beatV,&vele.tR0,vele.event,env.fprr); // register vtr beat
  fprintf(env.fplog,                           // log after VtrSense() service
    "%s#%d, Beat#%ld, RR=%.3lf, vtr.nA=%d, nextVV=%.3lf\n",
    vevt[vele.event],vele.cnt[vele.event],env.beatV,vele.RR,vtr.nA,pacer.NextVV);
}

/*********************************************************************
VtrPace: event service for ventricular pace
*********************************************************************/
int VtrPace()
{
  int status;
  char *capstat[] = {"LOC", "CAP"};

  LogEvent(env.fplog,"VtrPace"); 

  vele.event = E_VP;                           // event type: VP
  status = LOC;                                // default: loss of capture
  if (pacer.VP_AMP >= vele.VpThresh) {         // supra-threshold VP
     if (env.t-vele.tR0 >= vtr.ref) {          // ventrical non-refractory 
       status = CAP;                           // capture
       vtr.nR = StartWave(vtr.tR, vtr.nR, vtr.RetDly); // starts a retr vtr activation
       vele.RR = RegBeat(&env.beatV,&vele.tR0,vele.event,env.fprr); // register vtr beat
     }
     else                                      // ventricle refractory
       vtr.nB[RETROGRADE]++;                   // will not generate retr wave in vtr
  }     // otherwise, sub-threshold VP, will not generate retr wave in vtr
  vele.cnt[vele.event]++;                      // increment VP counter
  PacerInterval(vele.event);                   // update pacer measured intervals
  PacerTimerVx(pacer.MODE, vele.event);        // update pacer timers after VP
  
  RecordMarker(env.fpmk, vele.event, status);  // record ventricle event markers
  fprintf(env.fplog,                           // log after VtrPace() service
    "VP#%ld @%.3lf, ApTh %.3lf, %s, Beat#%ld, RR=%.3lf, vtr.nR=%d, nextVV=%.3lf\n", 
    vele.cnt[E_VP],pacer.VP_AMP,vele.VpThresh,capstat[status],env.beatV,vele.RR,
    vtr.nR,pacer.NextVV);

  return (status);
}

/*********************************************************************
PacerTimerVx: Update pacer timers after Vp or Vs
*********************************************************************/
void PacerTimerVx(int mode, int event)
{
  switch (event) {
    case E_VP:
      if (mode==DDD || mode==DDI) {                  // modes w/ VP & AS
        UpdateTimer(&pacer.tAB, SET, pacer.PVAB);    // start PVAB
        UpdateTimer(&pacer.tFFP, SET, pacer.FFPp);   // start FFPp
        UpdateTimer(&pacer.tPVARP, SET, pacer.PVARP);// start PVARP
      }
      if (mode==DDD || mode==DDI || mode==DVI) 
        UpdateTimer(&pacer.tUTI, SET, pacer.UTI);    // start UTI
      if (mode==DDD) {                               // A-based timing: DDD
        UpdateTimer(&pacer.tVP, OVW, -1);            // disable VP
        if (pacer.WKBseq>0) {                        // VP at the end of UTI
           if (pacer.WKBseq<=4)                      // long VA for <=4 WKB cycles
             pacer.NextVA = pacer.BI-MIN(pacer.SW, pacer.AVDp);
           else                                      // short VA after >4 WKB cycles
             pacer.NextVA = pacer.UTI-MIN(pacer.SW, pacer.AVDp);
           UpdateTimer(&pacer.tAP, SET, pacer.NextVA);// start VA timer for AP
        }
      }
      else if (mode==VOO)                            // V-based timing: VOO
        UpdateTimer(&pacer.tVP, SET, pacer.BI);      // start BI for VP
      else if (mode==DDI || mode==VVI || mode==DVI) {// V-based timing
        pacer.NextVV = GetVpIntv(vele.RR,pacer.BI,pacer.NextVV,pacer.VRS_METHOD);
        UpdateTimer(&pacer.tVP, SET, pacer.NextVV);  // start VV timer for VP
        if (mode==DDI || mode==DVI) {
          pacer.NextVA = pacer.NextVV-pacer.AVD;  
          UpdateTimer(&pacer.tAP, SET, pacer.NextVA);// start VA timer for AP        
        }
      }
      UpdateTimer(&pacer.tVB, SET, pacer.IPVB);      // start IPVB
      UpdateTimer(&pacer.tVRP, SET, pacer.VRP);      // start VRP
      break;
    case E_VS:
      if (pacer.tSW>0 && pacer.tSW>=pacer.tUTI)      // SW_VS (DDD, DDI, DVI only)
        UpdateTimer(&pacer.tVP, OVW, pacer.tSW);     // schedule SW_VP if no UTI violation
      else if (mode==DDD) {                          // A-based timing
        UpdateTimer(&pacer.tVP, OVW, -1);            // disable VP
        pacer.WKBseq = 0;                            // clear WKB sequence counter
      }
      else if (mode==DDI || mode==VVI || mode==DVI) {// V-based timing
        pacer.NextVV = GetVpIntv(vele.RR,pacer.BI,pacer.NextVV,pacer.VRS_METHOD);
        UpdateTimer(&pacer.tVP, SET, pacer.NextVV);  // start VV timer for VP
        if (mode==DDI || mode==DVI) {
          pacer.NextVA = pacer.NextVV-pacer.AVD;  
          UpdateTimer(&pacer.tAP, SET, pacer.NextVA);// start VA timer for AP        
        }
      }
      if (mode==DDD || mode==DDI || mode==DVI) 
        UpdateTimer(&pacer.tUTI, SET, pacer.UTI);    // start UTI
      UpdateTimer(&pacer.tVB, SET, pacer.ISVB);      // start ISVB
      UpdateTimer(&pacer.tVRP, SET, pacer.VRP);      // start VRP     
      if (mode==DDD || mode==DDI) {                  // modes w/ VP & AS
        UpdateTimer(&pacer.tFFP, SET, pacer.FFPs);   // start FFPs
      }
      break;
    case E_VES:
      UpdateTimer(&pacer.tVB, SET, pacer.ISVB);      // start ISVB
      UpdateTimer(&pacer.tVRP, SET, pacer.VRP);      // start VRP      
      if (mode==DDD || mode==DDI) {                  // modes w/ VP & AS
        UpdateTimer(&pacer.tFFP, SET, pacer.FFPs);   // start FFPs
        if (mode==DDD)                               // start PVARP w/ ext
          UpdateTimer(&pacer.tPVARP, SET, pacer.PVARP+pacer.PVARPext);
        else                                         // start PVARP w/o ext
          UpdateTimer(&pacer.tPVARP, SET, pacer.PVARP);
      }
      if (mode==DDD || mode==DDI || mode==DVI) 
        UpdateTimer(&pacer.tUTI, SET, pacer.UTI);    // start UTI
      if (mode==DDD) {                               // A-based timing
        UpdateTimer(&pacer.tVP, OVW, -1);            // disable VP
        UpdateTimer(&pacer.tAP, SET, pacer.BI);      // start BI for AP
        pacer.WKBseq = 0;                            // clear WKB sequence counter
      }
      else if (mode==DDI || mode==VVI) {             // V-based timing (no VES in DVI)
        pacer.NextVV = GetVpIntv(vele.RR,pacer.BI,pacer.NextVV,pacer.VRS_METHOD);
        UpdateTimer(&pacer.tVP, SET, pacer.NextVV);  // start VV timer for VP
        if (mode==DDI) {
          pacer.NextVA = pacer.NextVV-pacer.AVD;  
          UpdateTimer(&pacer.tAP, SET, pacer.NextVA);// start VA timer for AP        
        }
      }
      break;
    case E_VRS:
      UpdateTimer(&pacer.tVB, SET, pacer.ISVB);     // start ISVB
      break;
    default:                    // E_VSBLK
      break;                    // no action
  }
}

/*********************************************************************
PacerTimerAx: Update pacer timers after Ap or As
*********************************************************************/
void PacerTimerAx(int mode, int event)
{
  switch (event) {
    case E_AP:
      UpdateTimer(&pacer.tAB, SET, pacer.IPAB);   // start IPAB
      UpdateTimer(&pacer.tARP, SET, pacer.ARP);   // start ARP
      if (mode==DDD) pacer.AVD = pacer.AVDp;      // paced AV delay (DDD only)
      if (mode==DDD || mode==DDI || mode==DVI) {  // modes w/ AP & VS
        UpdateTimer(&pacer.tVB, SET, pacer.PAVB); // start PAVB  
        UpdateTimer(&pacer.tAVD, SET, pacer.AVD); // start AVD to check Vs
        UpdateTimer(&pacer.tSW, SET, pacer.SW);   // start SW
      }
      if (mode==DDD || mode==AAI || mode==AOO) {  // A-based timing
        UpdateTimer(&pacer.tAP, SET, pacer.BI);   // start BI for AP
        if (mode==DDD) PacerWkbAvd();             // check WKB and set AVD
      }
      else if (mode==DDI || mode==DVI)            // V-based timing
        UpdateTimer(&pacer.tAP, OVW, -1);         // disable AP
      break;
    case E_AS:
      UpdateTimer(&pacer.tAB, SET, pacer.ISAB);   // start ISAB
      UpdateTimer(&pacer.tARP, SET, pacer.ARP);   // start ARP
      if (mode==DDD) pacer.AVD = pacer.AVDs;      // sensed AV delay
      if (mode==DDD || mode==AAI) {               // A-based timing
        UpdateTimer(&pacer.tAP, SET, pacer.BI);   // start BI for AP
        if (mode==DDD) PacerWkbAvd();             // check WKB and set AVD
      }
      else if (mode==DDI)                         // V-based timing
        UpdateTimer(&pacer.tAP, OVW, -1);         // disable AP
      break;
    case E_ARS:
      UpdateTimer(&pacer.tAB, SET, pacer.ISAB);   // start ISAB
      break;
    case E_ASFFP:
      UpdateTimer(&pacer.tAB, SET, pacer.ISAB);   // start ISAB
      break;
    case E_ASPMT:
      UpdateTimer(&pacer.tAB, SET, pacer.ISAB);   // start ISAB
      UpdateTimer(&pacer.tARP, SET, pacer.ARP);   // start ARP
      break;
    default:                                      // E_ASBLK
      break;                                      // no action
  }
}

/*********************************************************************
PacerWkbAvd: Pacer checks WKB at UTI and sets appropriate AVD 
*********************************************************************/
void PacerWkbAvd()
{
  if (pacer.tUTI>pacer.AVD) {                     // AVD ends in UTI
    UpdateTimer(&pacer.tVP, SET, pacer.tUTI);     // shift VP to end of UTI
    pacer.WKBseq++;                               // inc WKB sequence count
  }
  else if (pacer.tUTI>0 && pacer.AVD>pacer.tUTI && pacer.WKBseq>0) {
    UpdateTimer(&pacer.tVP, SET, pacer.AVD);      // further shift VP to end of AVD
    pacer.WKBseq++;                               // inc WKB sequence count
  }
  else {                                          // standard (non WKB cycle)
    UpdateTimer(&pacer.tVP, SET, pacer.AVD);      // start AVD for VP
    pacer.WKBseq = 0;                             // clear WKB sequence counter
  }
  if (pacer.WKBseq>0)                             // log WKB sequence
    fprintf(env.fplog,"WKB#%d ",pacer.WKBseq);
}

/*********************************************************************
PacerInterval: Update pacer measured intervals after used A/V event
*********************************************************************/
void   PacerInterval(int event)
{
   if (event < NUM_VTYPE) {                  // V event
     if (event==E_VP || event==E_VS || event==E_VES) { // used V event
       pacer.LastUsedV = event;
       pacer.LastVV = pacer.tUsedV;          // measured VV 
       pacer.LastAV = pacer.tUsedA;          // measured AV (-1 if no LastUsedA)
       UpdateTimer(&pacer.tUsedV, OVW, 0);   // restart VA/VV measurement
     } // otherwise (Vrs, VsBlk) no update
   }
   else {                                    // A event
     event -= NUM_VTYPE;                     // restore A event enum value
     if (event==E_AP || event==E_AS || event==E_ASPMT) { // used A event
       pacer.LastUsedA = event;
       pacer.LastAA = pacer.tUsedA;          // measured AA 
       pacer.LastVA = pacer.tUsedV;          // measured VA (-1 if no LastUsedV)
       UpdateTimer(&pacer.tUsedA, OVW, 0);   // restart AV/AA measurement
     } // otherwise (Ars, AsFfp, AsBLK) no update
   }
}

/*********************************************************************
AtrSense: event service for atrial sense
args: type - sense type (INSTRINSIC, ECTOPIC, NOISE, CONDUCTED)
*********************************************************************/
void AtrSense(int type)
{
  char *aevt[] = {"AP", "AS", "ARS", "ASFFP", "ASPMT", "ASBLK"};
  
  LogEvent(env.fplog,"AtrSense"); 

  if (type == CONDUCTED)                // retrograde As: stop retrograde atr wave
    atr.nR = StopWave(atr.tR, atr.nR); 
  else if (type == NOISE) {             // noise As: increment noise AS counter
    arg.nNZ++;
    fprintf(env.fplog, "AtrNoise #%ld ", arg.nNZ);
  }                                     // otherwise intrinsic or ectopic As, no action
  
  if (pacer.MODE==VVI || pacer.MODE==AOO || pacer.MODE==VOO || pacer.MODE==DVI)  
    aele.event = E_ASBLK;                          // modes w/o AS, treat as AsBlk
  else {                                           // modes DDD, DDI, AAI
    if (pacer.tAB > 0) aele.event = E_ASBLK;       // AS in blanking window
    else {                                         // unblanked incl. As, Ars, AsFfp, AsPmt
      if (pacer.tFFP > 0) aele.event = E_ASFFP;    // AS in FF protection window (DDD & DDI)
      else if (pacer.tARP > 0) aele.event = E_ARS; // AS in atrial refr window 
      else {                                       // AsPmt or As 
        if (pacer.tPVARP > 0) {                    // AS in PVARP window (DDD & DDI)
          aele.event = E_ASPMT;
        }
        else {
          aele.event = E_AS;                       // tracked AS
        }
      }
    }
  }
  aele.cnt[aele.event]++;   // increment respective atrial sense counter
  PacerInterval(aele.event+NUM_VTYPE);           // update pacer measured intervals
  PacerTimerAx(pacer.MODE, aele.event);          // update pacer timers after AS
  
  RecordMarker(env.fpmk, aele.event+NUM_VTYPE, type); // record atrial event markers
  if (type != NOISE) 
    aele.PP = RegBeat(&env.beatA,&aele.tA0,aele.event,env.fpaa); // register atr beat
  fprintf(env.fplog,        // log after AtrSense() service
    "%s#%d, atr.nR=%ld\n",aevt[aele.event],aele.cnt[aele.event],atr.nR);
}

/*********************************************************************
AtrPace: event service for atrial pace
*********************************************************************/
int AtrPace()
{
  int status;
  char *capstat[] = {"LOC", "CAP"};
  
  LogEvent(env.fplog,"AtrPace"); 

  aele.event = E_AP;                  // event marker: AP
  status = LOC;                       // default: loss of capture
  if (pacer.AP_AMP >= aele.ApThresh) {// supra-threshold AP
     if (env.t-aele.tA0 >= atr.ref) { // atrial non-refractory 
       status = CAP;                  // capture
       arg.dV = avj.Vt-avj.Vr;        // equivalent to strong atrial impulse
       atr.nA = StartWave(atr.tA, atr.nA, atr.AntDly); // starts ante activation (no sense compens.)
       aele.PP = RegBeat(&env.beatA,&aele.tA0,aele.event,env.fpaa); // register atr beat
     }
     else                             // atrial refractory
       atr.nB[ANTEGRADE]++;           // will not generate ante wave in atr
  }     // otherwise, sub-threshold Ap, will not generate ante wave in atr
  aele.cnt[aele.event]++;   // increment AP counter
  PacerInterval(aele.event+NUM_VTYPE);   // update pacer measured intervals
  PacerTimerAx(pacer.MODE, aele.event);  // update pacer timers after AP
  
  RecordMarker(env.fpmk, aele.event+NUM_VTYPE, status); // record atrial event markers
  fprintf(env.fplog,
    "AP#%ld @%.3lf, ApTh %.3lf, %s\n", // log after AtrPace() service
    aele.cnt[E_AP],pacer.AP_AMP, aele.ApThresh, capstat[status]);
    
  return (status);
}

/*********************************************************************
GetNextIntv: get next ARG or VRG output interval
arg:  chamber - ATRIUM or VENTRICLE
      type - point to arg.type or vrg.type
*********************************************************************/
double GetNextIntv(int chamber, int *type)
{
  int model;
  long r, nAA;
  double intv, mean, std, ref, EB_Prob, EB_CI1, EB_CI2;
  
  if (chamber==ATRIUM) {
    model = arg.MODEL;
    mean = 1/arg.lambda;
    std = arg.AAstd;
    ref = atr.ref;
    EB_Prob = arg.EB_Prob;
    EB_CI1 = arg.EB_CI1;
    EB_CI2 = arg.EB_CI2;
    nAA = arg.nAA[REFRACTORY]+arg.nAA[PHASE4];
  }
  else {
    model = vrg.MODEL;
    mean = vrg.mean;
    std = vrg.std;
    ref = vtr.ref;
    EB_Prob = vrg.EB_Prob;
    EB_CI1 = vrg.EB_CI1;
    EB_CI2 = vrg.EB_CI2;
    if (model>2) model = 0; // default 
  }
  switch (model) {
    case 1:      // random value from truncated exponential distribution
      r = (long) random();
      intv = ref+expdev(mean-ref,&r); // limit intv >= ref 
      break;
    case 2:      // random value from uniform distribution
      intv = mean+std*mean*unirand();
      break;
    case 3:      // random value from Gaussian distribution
      intv = gaussian(mean,std*mean,unirand(),unirand());
      break;
    case 4:      // atrial pacing protocl with programmed S1S2
      if (nAA < env.MAX_BEAT/2) intv = mean;
      else  intv = arg.S1S2;
      break;
    case 5:      // atrial pacing protocl with programmed S1S2/S2S3
      if (nAA < env.MAX_BEAT/2) intv = mean;
      else  {  
        if (nAA%2==0) intv = arg.S1S2; // even beat
        else intv = arg.S2S3;          // odd beat
      }
      break;
    case 6:      // read from external AA intervals
      if (nAA>MAX_AA) nAA = 0;
      intv = aaintv[nAA];
      break;
    default:     // fixed rate, flutter protocol
      intv = mean;
      break;
   }
  if (intv < ref) intv = ref; // ensure intv >= ref

  *type = INTRINSIC;          // default output type INTRINSIC
  if (unirand() < EB_Prob) {  // possible EB in this cycle
    mean = EB_CI1+unirand()*(EB_CI2-EB_CI1); // EB coupling interval
    if (mean < intv) {
       *type = ECTOPIC;      // next impulse will be ectopic beat
       intv = mean;          // update intv to EB coupling interval
    } // otherwise, EB will be inhibited due to earlier normal depol.
  } // otherwise, EB not possible in this cycle
  return intv;
}

/*********************************************************************
GetVpIntv: get next ventricular pace interval
*********************************************************************/

double GetVpIntv(double RR,double BI,double VpIntv, int method)
{
 double NextVpIntv;

 // apply appropriate ventricular rate smoothing algorithms below
 switch (method) {
    case 0:
       NextVpIntv = BI;
       break;
//    case 1:
//       NextVpIntv = adaptive_VRS(RR,VpIntv);
//       break;
//    case 2:
//       NextVpIntv = wittkampf_VRS(RR,VpIntv);
//       break;
    default:
       NextVpIntv = BI;
       break;
 }

 return NextVpIntv;
}
