/****************************************************************
  idhp.h

  Header file for the Integrade Dual-chamber Heart & Pacer model
  Reference: 
     J Lian, D Muessig, "Heart rhythm and cardiac pacing: an 
     integrated dual-chamber heart and pacer model." 
     Ann Biomed Egn. 37(1):64-81, 2009. 

  Author: Jie Lian
*****************************************************************/
#ifndef _IDHP_H_
#define _IDHP_H_

#define MAX_TM    2   // max number of waves allowed in atr, vtr, or AVJ
#define BUFLIM    160 // string buffer length

enum {ATRIUM, VENTRICLE};                // chamber
enum {ANTEGRADE,RETROGRADE,BIDIRECTION}; // direction of activation
enum {REFRACTORY,PHASE4};                // AVJ phases
enum {LOC, CAP};                         // capture status: loss-of-capture & capture
enum {OVW, SET, INC, DEC};               // actions for pacer timer update
enum {AOO, VOO, AAI, VVI, DVI, DDI, DDD};// pacer mode
enum {INTRINSIC, ECTOPIC, NOISE, CONDUCTED}; // true sense types
enum {E_VP, E_VS, E_VES, E_VRS, E_VSBLK};            // vele event type
enum {E_AP, E_AS, E_ARS, E_ASFFP, E_ASPMT, E_ASBLK}; // aele event type

#define NUM_MODE  (DDD+1)     // number of pacer modes
#define NUM_VTYPE (E_VSBLK+1) // number of ventricle event types (5)
#define NUM_ATYPE (E_ASBLK+1) // number of atrial event types (6)

typedef struct {      // data structure of simulation environment
  // programmable parameters
  char fnRR[BUFLIM];  // output RR interval filename
  char fnAA[BUFLIM];  // output AA interval filename
  char fnAV[BUFLIM];  // output AVJ status filename
  char fnMK[BUFLIM];  // output event marker filename
  char fnLOG[BUFLIM]; // output event log filename
  long MAX_BEAT;      // max RR intervals to run
  double MAX_TIME;    // max running time in seconds
  double Ts;          // sampling interval
  double RR0;         // initial RR interval
  // working parameters
  FILE *fprr;         // point to fnRR
  FILE *fpaa;         // point to fnAA
  FILE *fpav;         // point to fnAV
  FILE *fpmk;         // point to fnMK
  FILE *fplog;        // point to fnLog
  long beatA;         // atrial beat counter 
  long beatV;         // ventricle beat counter  
  double t;           // simulation time [0 .. MAX_TIME]
} EnvStruct;

typedef struct {      // data structure of Atrial Rhythm Generator model
  // programmable parameters
  int    MODEL;       // ARG interval generator method
  double lambda;      // mean rate of Afib bombardment on AV junction
  double AAstd;       // standard deviation of AA interval
  double dVmean;      // mean potential increment by Afib bombardment
  double dVstd;       // standard deviation of dV
  double S1S2;        // atrial pacing protocol S1S2 interval
  double S2S3;        // atrial pacing protocol S2S3 interval
  double EB_Prob;     // probability of ectopic beat AES (per beat) [0..1]
  double EB_CI1;      // shortest possible AES coupling interval
  double EB_CI2;      // longest possible AES coupling interval
  double NZ_Prob;     // probability of atr noise (per sample) [0..1]
  // working parameters
  double dV;          // actual calculated dV
  double t;           // count-down timer for ARG output
  int    type;        // ARG output type: INSTRINSIC, ECTOPIC, NOISE
  long   nNZ;         // number of atr noise events
  long   nEB;         // number of AES
  long   nAA[2];      // number of atrial impulses in AVJ phase 4 & refr period
  long   nRS[2];      // number of ARG resets
} ArgStruct;

typedef struct {      // data structure of Ventricle Rhythm Generator model
  // programmable parameters
  int    MODEL;       // VRG interval generator method
  double mean;        // mean VRG output interval
  double std;         // standard deviation of VRG output interval
  double EB_Prob;     // probability of ectopic beat VES (per beat) [0..1]
  double EB_CI1;      // shortest possible VES coupling interval
  double EB_CI2;      // longest possible VES coupling interval
  double NZ_Prob;     // probability of vtr noise (per sample) [0..1]
  // working parameters
  double t;           // count-down timer for VRG output
  int    type;        // VRG output type: INSTRINSIC, ECTOPIC, NOISE
  long   nNZ;         // number of vtr noise events
  long   nEB;         // number of VES
} VrgStruct;

typedef struct {      // data structure of atrial model
  // programmable parameters
  double AntDly;      // ante conduction delay from atr source to AVJ
  double RetDly;      // retr conduction delay from AVJ to atr source
  double ref;         // atrial refractory period
  // working parameters
  double tA[MAX_TM];  // timer starts after ante atr depolarization
  double tR[MAX_TM];  // timer starts after retr atr depolarization
  int    nA;          // number of active antegrade conduction timers
  int    nR;          // number of active retrograde conduction timers
  double tA0;         // time of last ante atr depol, start ante cond
  double tR0;         // time of last retr atr depol, start retr cond
  long   nF[2];       // number of atrial fusions in AVJ phase 4 & refr period
  long   nB[2];       // number of blocked impulses (ante & retr)
} AtrStruct;

typedef struct {      // data structure of AV junction model
  // programmable parameters
  double Vt;          // AV junction threshold potential
  double Vr;          // AV junction resting potential
  double dVdt;        // original V4 (AVJ spontaneous depolarization slope)
  double MinAVDa;     // minimum anterograde AV conduction delay
  double MinAVDr;     // minimum retrograde AV conduction delay
  double MinRef;      // minimum AVJ refractory period
  double Ref_std;     // standard deviation of the AVJ refractory period
  double alpha;       // max extension of AV conduction time
  double beta;        // max extension of AVJ refractory period
  double tau_c;       // time constant for AV conduction curve
  double tau_r;       // time constant for AVJ recovery curve
  double delta;       // electrotonic modulation control parameter (strength)
  double theta;       // electrotonic modulation control parameter (time)
  // working parameters
  int    phase;       // AVJ phase flag (phase4 or refractory)
  int    dir;         // direction of most recent activation
  double Vm;          // AVJ membrane potential
  double V4;          // actual AVJ spontaneous depolarization slope
  double ref;         // actually calculated refractory period
  double AntDly;      // corrected antegrade AV delay
  double RetDly;      // corrected retrograde AV delay
  double RT;          // AVJ recovery time
  double RT0;         // time at the start of AVJ recovery period
  double tA[MAX_TM];  // AVJ ante activation timer 
  double tR[MAX_TM];  // AVJ retr activation timer 
  double tA0;         // time of last ante AVD timeout, start (ante) vtr refr 
  double tR0;         // time of last retr AVD timeout, start (retr) atr refr 
  int    nA;          // number of active ante. AV conduction timers
  int    nR;          // number of active retr. AV conduction timers
  double t;           // count-up timer started after AVJ enters refractory period
  long   nF[2];       // number of AVJ fusions in phase 4 & refr period
  long   nB[2];       // number of blocked impulses (ante & retr)
} AvjStruct;

typedef struct {      // data structure of ventricle model
  // programmable parameters
  double AntDly;      // anterograde conduction delay from AVJ to RV electrode
  double RetDly;      // retrograde conduction delay from RV electrode to AVJ
  double ref;         // ventricular refractory period
  // working parameters
  double tA[MAX_TM];  // timer starts after AV depol, ante cond for Vs
  double tR[MAX_TM];  // timer starts after Vp for retr conduction
  int    nA;          // number of active retrograde conduction timers
  int    nR;          // number of active antegrade conduction timers
  long   nF[2];       // number of ventricular fusions in phase 4 & refr period
  long   nB[2];       // number of blocked impulses (ante & retr)
} VtrStruct;

typedef struct {      // data structure of RV electrode model
  // programmable parameters
  double VpThresh;    // ventricular pacing threshold
//  double VpDly;       // delay of Vp evoked response
  // working parameters
  int    event;       // ventricular event type
  double RR;          // measured RR interval
  double tR0;         // time of last Vp or Vs, start (retr) vtr refr period
  long   cnt[5];      // ventricle event counters
} VeleStruct;

typedef struct {      // data structure of RA electrode model
  // programmable parameters
  double ApThresh;    // atrial pacing threshold
  // working parameters
  int    event;       // atrial event type 
  double PP;          // measured PP interval
  double tA0;         // time of last Ap or As, start (ante) atr refr period
  long   cnt[6];      // atrial event counters
} AeleStruct;

typedef struct {      // data structure of pacer model
  // programmable parameters
  int    MODE;        // pacing mode
  double BI;          // basic interval
  double AVDp;        // programmed paced AV delay 
  double AVDs;        // programmed sensed AV delay (must <=AVDp)
  double PVARP;       // post-ventricular atrial refractory period (325ms)
  double PVARPext;    // PVARP extension after VES (150ms, DDD)
  double FFPp;        // far-field protection window after VP (150ms)
  double FFPs;        // far-field protection window after VS (100ms)
  double ARP;         // atrial refractory period (300ms, must >=225ms)
  double VRP;         // ventricular refractory period (250ms)
  double PVAB;        // post-VP atr blanking period (x-channel blanking 30ms)
  double PAVB;        // post-AP vtr blanking period (x-channel blanking 30ms)
  double IPAB;        // post-AP detection holdoff period (in-channel AP blanking 120ms)
  double IPVB;        // post-VP detection holdoff period (in-channel VP blanking 250ms)
  double ISAB;        // post-AS detection holdoff period (in-channel AS blanking 101ms)
  double ISVB;        // post-VS detection holdoff period (in-channel VS blanking 101ms)
  double UTI;         // upper tracking interval
  double SW;          // safety window (100ms) 
  double AP_AMP;      // atrial pacing amplitude
  double VP_AMP;      // ventricular pacing amplitude
  int    VRS_METHOD;  // ventricular rate smoothing method
  // working parameters
  double tPVARP;      // count-down timer for PVARP
  double tFFP;        // count-down timer for FFP (FFPp or FFPs)
  double tARP;        // count-down timer for ARP
  double tVRP;        // count-down timer for VRP
  double tAB;         // count-down timer for AB (pAB or PVAB)
  double tVB;         // count-down timer for VB (pVB or PAVB)
  double tUTI;        // count-down timer for UTI
  double tAP;         // count-down timer for AP
  double tVP;         // count-down timer for VP
  double tAVD;        // count-down timer for AVD
  double tSW;         // count-donw timer for SW
  double tUsedA;      // count-up timer for AA and AV interval measurement
  double tUsedV;      // count-up timer for VA and VV interval measurement
  double AVD;         // operating AV delay 
  int    WKBseq;      // sequence of Wenckebach behavior
  int    LastUsedA;   // last used atrial event type (Ap, As, AsPmt)
  int    LastUsedV;   // last used ventricle event type (Vp, Vs, VES)
  double LastAA;      // last used AA interval 
  double LastVA;      // last used VA interval
  double LastAV;      // last used AV interval
  double LastVV;      // last used VV interval
  double NextVV;      // ventricular pacing interval (escape interval)
  double NextVA;      // VA interval (ventricle-based timing)
} PacerStruct;

// functions defined by IDHP model

double expdev(double, long *);
double gaussian(double, double, double, double);
double ran2(long *);
double ran3(int *);
double unirand(void);

void   ErrMsg(char *, int);
void   ReadConfig(char *fn);
double ReadLine(FILE *, double, char *);
void   LogConfig(FILE *);
void   LogEvent(FILE *, char *);
void   LogStat(FILE *);
void   RecordMarker(FILE *, int, int);

void   ModelInit(void);
void   PacerInit(void);
void   ModelExit(void);
void   UpdateAtTs(void);
void   UpdateTimer(double *, int, double);

void   AnteHitAvj(void);
void   RetrHitAvj(void);
void   AnteEscAvj(void);
void   RetrEscAvj(void);
void   ActivateAvj(void);
void   StartAvjRef(void);
void   StartAvjPh4(void);

int    StartWave(double *, int, double);
int    StopWave(double *, int);

int    ArgOutput(void);
void   ArgReset(void);
void   AtrFusion(void);
void   AtrSense(int);
int    AtrPace(void);

int    VrgOutput(void);
void   VrgReset(void);
void   VtrFusion(void);
void   VtrSense(int);
int    VtrPace(void);

void   PacerTimerAx(int, int);
void   PacerTimerVx(int, int);
void   PacerWkbAvd(void);
void   PacerInterval(int);

double RegBeat(long *, double *, int, FILE *);
double GetVpIntv(double, double, double, int);
double GetNextIntv(int, int *);
long   ReadAAintv(double *);

#endif  // #ifndef _IDHP_H_
